const $ = require('../../core/renderer');
const ko = require('knockout');
const Callbacks = require('../../core/utils/callbacks');
const errors = require('../../core/errors');
const inflector = require('../../core/utils/inflector');
const isPlainObject = require('../../core/utils/type').isPlainObject;
const registerComponentCallbacks = require('../../core/component_registrator_callbacks');
const Widget = require('../../ui/widget/ui.widget');
const KoTemplate = require('./template').KoTemplate;
const Editor = require('../../ui/editor/editor');
const Locker = require('../../core/utils/locker');
const getClosestNodeWithContext = require('./utils').getClosestNodeWithContext;
const config = require('../../core/config');

const LOCKS_DATA_KEY = 'dxKoLocks';
const CREATED_WITH_KO_DATA_KEY = 'dxKoCreation';

const editorsBindingHandlers = [];

const registerComponentKoBinding = function(componentName, componentClass) {

    if(componentClass.subclassOf(Editor)) {
        editorsBindingHandlers.push(componentName);
    }

    ko.bindingHandlers[componentName] = {
        init: function(domNode, valueAccessor) {
            const $element = $(domNode);
            const optionChangedCallbacks = Callbacks();
            let optionsByReference = {};
            let component;
            const knockoutConfig = config().knockout;
            const isBindingPropertyPredicateName = knockoutConfig && knockoutConfig.isBindingPropertyPredicateName;
            let isBindingPropertyPredicate;
            let ctorOptions = {
                onInitializing: function() {
                    optionsByReference = this._getOptionsByReference();

                    ko.computed(function() {
                        const model = ko.unwrap(valueAccessor());

                        if(component) {
                            component.beginUpdate();
                        }

                        isBindingPropertyPredicate = isBindingPropertyPredicateName && model && model[isBindingPropertyPredicateName];

                        unwrapModel(model);

                        if(component) {
                            component.endUpdate();
                        }

                    }, null, { disposeWhenNodeIsRemoved: domNode });

                    component = this;
                },
                modelByElement: function($element) {
                    if($element.length) {
                        const node = getClosestNodeWithContext($element.get(0));
                        return ko.dataFor(node);
                    }
                },
                nestedComponentOptions: function(component) {
                    return {
                        modelByElement: component.option('modelByElement'),
                        nestedComponentOptions: component.option('nestedComponentOptions')
                    };
                },
                _optionChangedCallbacks: optionChangedCallbacks,
                integrationOptions: {
                    watchMethod: function(fn, callback, options) {
                        options = options || {};

                        let skipCallback = options.skipImmediate;
                        const watcher = ko.computed(function() {
                            const newValue = ko.unwrap(fn());
                            if(!skipCallback) {
                                callback(newValue);
                            }
                            skipCallback = false;
                        });
                        return function() {
                            watcher.dispose();
                        };
                    },
                    templates: {
                        'dx-polymorph-widget': {
                            render: function(options) {
                                let widgetName = ko.utils.unwrapObservable(options.model.widget);
                                if(!widgetName) {
                                    return;
                                }

                                if(widgetName === 'button' || widgetName === 'tabs' || widgetName === 'dropDownMenu') {
                                    const deprecatedName = widgetName;
                                    widgetName = inflector.camelize('dx-' + widgetName);
                                    errors.log('W0001', 'dxToolbar - \'widget\' item field', deprecatedName, '16.1', 'Use: \'' + widgetName + '\' instead');
                                }

                                const markup = $('<div>').attr('data-bind', widgetName + ': options').get(0);
                                $(options.container).append(markup);
                                ko.applyBindings(options.model, markup);
                            }
                        }
                    },
                    createTemplate: function(element) {
                        return new KoTemplate(element);
                    }
                }
            };
            const optionNameToModelMap = {};

            const applyModelValueToOption = function(optionName, modelValue, unwrap) {
                const locks = $element.data(LOCKS_DATA_KEY);
                const optionValue = unwrap ? ko.unwrap(modelValue) : modelValue;

                if(ko.isWriteableObservable(modelValue)) {
                    optionNameToModelMap[optionName] = modelValue;
                }

                if(component) {
                    if(locks.locked(optionName)) {
                        return;
                    }

                    locks.obtain(optionName);

                    try {
                        if(ko.ignoreDependencies) {
                            ko.ignoreDependencies(component.option, component, [optionName, optionValue]);
                        } else {
                            component.option(optionName, optionValue);
                        }
                    } finally {
                        locks.release(optionName);
                    }
                } else {
                    ctorOptions[optionName] = optionValue;
                }
            };

            const handleOptionChanged = function(args) {
                const optionName = args.fullName;
                const optionValue = args.value;

                if(!(optionName in optionNameToModelMap)) {
                    return;
                }

                const $element = this._$element;
                const locks = $element.data(LOCKS_DATA_KEY);

                if(locks.locked(optionName)) {
                    return;
                }

                locks.obtain(optionName);
                try {
                    optionNameToModelMap[optionName](optionValue);
                } finally {
                    locks.release(optionName);
                }
            };

            const createComponent = function() {
                optionChangedCallbacks.add(handleOptionChanged);
                $element
                    .data(CREATED_WITH_KO_DATA_KEY, true)
                    .data(LOCKS_DATA_KEY, new Locker());

                new componentClass($element, ctorOptions);

                ctorOptions = null;
            };

            const unwrapModelValue = function(currentModel, propertyName, propertyPath) {
                if(propertyPath === isBindingPropertyPredicateName) {
                    return;
                }

                if(
                    !isBindingPropertyPredicate ||
                    isBindingPropertyPredicate(propertyPath, propertyName, currentModel)
                ) {
                    let unwrappedPropertyValue;

                    ko.computed(function() {
                        const propertyValue = currentModel[propertyName];
                        applyModelValueToOption(propertyPath, propertyValue, true);
                        unwrappedPropertyValue = ko.unwrap(propertyValue);
                    }, null, { disposeWhenNodeIsRemoved: domNode });

                    if(isPlainObject(unwrappedPropertyValue)) {
                        if(!optionsByReference[propertyPath]) {
                            unwrapModel(unwrappedPropertyValue, propertyPath);
                        }
                    }
                } else {
                    applyModelValueToOption(propertyPath, currentModel[propertyName], false);
                }
            };

            var unwrapModel = function(model, propertyPath) {
                for(const propertyName in model) {
                    if(Object.prototype.hasOwnProperty.call(model, propertyName)) {
                        unwrapModelValue(model, propertyName, propertyPath ? [propertyPath, propertyName].join('.') : propertyName);
                    }
                }
            };

            createComponent();

            return {
                controlsDescendantBindings: componentClass.subclassOf(Widget)
            };
        }
    };

    if(componentName === 'dxValidator') {
        ko.bindingHandlers['dxValidator'].after = editorsBindingHandlers;
    }
};

registerComponentCallbacks.add(function(name, componentClass) {

    registerComponentKoBinding(name, componentClass);

});
